from pathlib import Path
from typing import List

from py_app_dev.core.logging import logger

from yanga.ybuild.pipeline import Stage


class MyStage(Stage):
    def run(self) -> int:
        logger.info(f"Running my stage with output dir '{self.output_dir}'")
        return 0

    def get_name(self) -> str:
        return "my_stage"

    def get_inputs(self) -> List[Path]:
        return []

    def get_outputs(self) -> List[Path]:
        return []



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List


class IncludeDirectoriesProvider(ABC):
    @abstractmethod
    def get_include_directories(self) -> List[Path]:
        ...



from pathlib import Path
from typing import List

from kspl.generate import HeaderWriter
from kspl.kconfig import KConfig
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger
from py_app_dev.core.scoop_wrapper import ScoopWrapper

from yanga.ybuild.generators.build_system import (
    BuildSystemBackend,
    BuildSystemGenerator,
)
from yanga.ybuild.include_directories_provider import IncludeDirectoriesProvider

from .backends.cmake import CMakeRunner
from .backends.generated_file import GeneratedFile
from .environment import BuildEnvironment
from .pipeline import Stage


class YangaVEnvInstall(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()

    @property
    def install_dirs(self) -> List[Path]:
        return [
            self.project_root_dir / dir
            for dir in [".venv/Scripts", ".venv/bin"]
            if (self.project_root_dir / dir).exists()
        ]

    def get_name(self) -> str:
        return "yanga_venv_install"

    def run(self) -> int:
        if not self.install_dirs:
            self.logger.warning("No virtual environment found. Attempt to run 'bootstrap.py'.")
            build_script_path = self.project_root_dir / "bootstrap.py"
            if not build_script_path.exists():
                raise UserNotificationException(
                    "Failed to find bootstrap script. Make sure that the project is initialized correctly."
                )
            self.environment.create_process_executor(
                ["python", build_script_path.as_posix()],
                cwd=self.project_root_dir,
            ).execute()
        else:
            self.logger.info("Found virtual environment. Skip 'bootstrap.py'.")
        self.environment.add_install_dirs(self.install_dirs)
        return 0

    def get_inputs(self) -> List[Path]:
        return []

    def get_outputs(self) -> List[Path]:
        return []


class YangaScoopInstall(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()
        self.install_dirs: List[Path] = []

    def get_name(self) -> str:
        return "yanga_scoop_install"

    @property
    def scoop_file(self) -> Path:
        return self.project_root_dir.joinpath("scoopfile.json")

    def run(self) -> int:
        self.logger.info(f"Run {self.__class__.__name__} stage. Output dir: {self.output_dir}")
        installed_apps = ScoopWrapper().install(self.scoop_file)
        for app in installed_apps:
            self.install_dirs.extend(app.get_all_required_paths())
        # Update the install directories for the subsequent stages
        self.environment.add_install_dirs(list(set(self.install_dirs)))
        return 0

    def get_inputs(self) -> List[Path]:
        return [self.scoop_file]

    def get_outputs(self) -> List[Path]:
        return self.install_dirs


class YangaWestInstall(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()
        self.artifacts_locator = self.environment.artifacts_locator

    def get_name(self) -> str:
        return "yanga_west_install"

    @property
    def west_manifest_file(self) -> Path:
        return self.project_root_dir.joinpath("west.yaml")

    def run(self) -> int:
        self.logger.info(f"Running {self.__class__.__name__} stage. Output dir: {self.output_dir}")
        try:
            self.environment.create_process_executor(
                [
                    "west",
                    "init",
                    "-l",
                    "--mf",
                    self.west_manifest_file.as_posix(),
                    self.project_root_dir.joinpath("build/west").as_posix(),
                ],
                cwd=self.project_root_dir,
            ).execute()
            self.environment.create_process_executor(
                ["west", "update"],
                cwd=self.project_root_dir.joinpath("build"),
            ).execute()
        except Exception as e:
            raise UserNotificationException(f"Failed to initialize and update with west: {e}")

        return 0

    def get_inputs(self) -> List[Path]:
        return [self.west_manifest_file]

    def get_outputs(self) -> List[Path]:
        return []


class YangaBuildConfigure(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()
        self.generated_files: List[GeneratedFile] = []

    def get_name(self) -> str:
        return "yanga_build_configure"

    def run(self) -> int:
        self.logger.info(f"Run {self.__class__.__name__} stage. Output dir: {self.output_dir}")
        self.generated_files = BuildSystemGenerator(
            BuildSystemBackend.CMAKE, self.environment, self.output_dir
        ).generate()
        for file in self.generated_files:
            file.to_file()
        return 0

    def get_inputs(self) -> List[Path]:
        return self.environment.user_config_files

    def get_outputs(self) -> List[Path]:
        return [file.path for file in self.generated_files]


class YangaBuildRun(Stage):
    """Run CMake and build the project.
    This stage is always executed. The dependencies are handled by CMake."""

    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()

    def get_name(self) -> str:
        return "yanga_build_run"

    def run(self) -> int:
        self.logger.info(f"Run {self.__class__.__name__} stage. Output dir: {self.output_dir}")
        CMakeRunner(self.environment.install_dirs).run(self.output_dir, self.environment.build_request.target_name)
        return 0

    def get_inputs(self) -> List[Path]:
        return []

    def get_outputs(self) -> List[Path]:
        return []


class KConfigIncludeDirectoriesProvider(IncludeDirectoriesProvider):
    def __init__(self, output_dir: Path) -> None:
        self.output_dir = output_dir

    def get_include_directories(self) -> List[Path]:
        return [self.output_dir]


class YangaKConfigGen(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()
        self.input_files: List[Path] = []

    def get_name(self) -> str:
        return "yanga_kconfig_gen"

    @property
    def header_file(self) -> Path:
        return self.output_dir.joinpath("autoconf.h")

    def run(self) -> int:
        self.logger.info(f"Run {self.__class__.__name__} stage. Output dir: {self.output_dir}")
        kconfig_model_file = self.project_root_dir.joinpath("KConfig")
        if not kconfig_model_file.is_file():
            self.logger.info("No KConfig file found. Skip this stage.")
            return 0
        kconfig = KConfig(
            kconfig_model_file,
            self.environment.config_file,
        )
        self.input_files = kconfig.get_parsed_files()
        config = kconfig.collect_config_data()
        HeaderWriter(self.header_file).write(config)
        # Update the include directories for the subsequent stages
        self.environment.add_include_dirs_provider(KConfigIncludeDirectoriesProvider(self.output_dir))
        return 0

    def get_inputs(self) -> List[Path]:
        # TODO: Use as input only the user config file where variant configuration is defined.
        # Now all the user config files are used as inputs, which will trigger the generation
        # if any of the file has changed.
        return self.environment.user_config_files + self.input_files

    def get_outputs(self) -> List[Path]:
        return [self.header_file]



from enum import Enum, auto
from pathlib import Path
from typing import Any, List

from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger

from yanga.ybuild.backends.cmake import (
    CMakeAddExecutable,
    CMakeAddSubdirectory,
    CMakeCommand,
    CMakeComment,
    CMakeContent,
    CMakeCustomCommand,
    CMakeCustomTarget,
    CMakeEmptyLine,
    CMakeEnableTesting,
    CMakeExecuteProcess,
    CMakeFile,
    CMakeInclude,
    CMakeIncludeDirectories,
    CMakeListAppend,
    CMakeMinimumVersion,
    CMakeObjectLibrary,
    CMakePath,
    CMakeProject,
    CMakeVariable,
)
from yanga.ybuild.components import BuildComponent
from yanga.ybuild.generators.build_system_request import (
    BuildVariantRequest,
    CompileComponentRequest,
    TestComponentRequest,
    TestVariantRequest,
)

from ..backends.generated_file import GeneratedFile
from ..environment import BuildEnvironment


def make_list_unique(seq: List[Any]) -> List[Any]:
    return list(dict.fromkeys(seq))


class BuildSystemBackend(Enum):
    CMAKE = auto()
    NINJA = auto()


class BuildComponentAnalyzer:
    def __init__(self, components: List[BuildComponent]) -> None:
        self.components = components

    def collect_sources(self) -> List[Path]:
        files: List[Path] = []
        for component in self.components:
            files.extend([component.path.joinpath(source) for source in component.sources])
        return files

    def collect_test_sources(self) -> List[Path]:
        files: List[Path] = []
        for component in self.components:
            files.extend([component.path.joinpath(source) for source in component.test_sources])
        return files

    def collect_include_directories(self) -> List[Path]:
        # TODO: check if there are specific include directories for each component
        includes = [path.parent for path in self.collect_sources()]
        # remove duplicates and return
        return make_list_unique(includes)

    def get_testable_components(self) -> List[BuildComponent]:
        return [component for component in self.components if component.test_sources]

    def is_testable(self) -> bool:
        return any(component.test_sources for component in self.components)


class BuildSystemGenerator:
    def __init__(
        self,
        backend: BuildSystemBackend,
        environment: BuildEnvironment,
        output_dir: Path,
    ):
        self.logger = logger.bind()
        self.backend = backend
        self.environment = environment
        self.output_dir = output_dir
        self.output_test_dir = output_dir.joinpath("test")

    def generate(self) -> List[GeneratedFile]:
        files = []
        if self.backend == BuildSystemBackend.CMAKE:
            files.extend(CmakeBuildFilesGenerator(self.environment, self.output_dir).generate())
            if BuildComponentAnalyzer(self.environment.components).is_testable():
                self.logger.info("Generating test build files")
                files.extend(CmakeTestFilesGenerator(self.environment, self.output_test_dir).generate())
        else:
            raise UserNotificationException("Only CMake build system files generation is supported for now")
        # TODO: CMakeFile objects are GeneratedFile objects. Try to fix warning without ignoring type check
        return files  # type: ignore


class CmakeBuildFilesGenerator:
    """Generates CMakeLists.txt, variant.cmake, and components.cmake files
    for building the variant for a specific platform.
    """

    def __init__(self, environment: BuildEnvironment, output_dir: Path) -> None:
        self.environment = environment
        self.output_dir = output_dir
        # The directory where the CMakeLists.txt file is located
        self.cmake_current_list_dir = CMakePath(self.output_dir, "CMAKE_CURRENT_LIST_DIR")
        # The directory where the build files will be generated
        self.cmake_build_dir = CMakePath(self.output_dir / "build", "CMAKE_BUILD_DIR")

    @property
    def variant_name(self) -> str:
        return self.environment.variant_name

    @property
    def variant_cmake_file(self) -> CMakePath:
        return self.cmake_current_list_dir.joinpath("variant.cmake")

    @property
    def components_cmake_file(self) -> CMakePath:
        return self.cmake_current_list_dir.joinpath("components.cmake")

    def generate(self) -> List[CMakeFile]:
        return [self.create_cmake_lists(), self.create_variants_cmake(), self.create_components_cmake()]

    def create_cmake_lists(self) -> CMakeFile:
        cmake_file = CMakeFile(self.output_dir.joinpath("CMakeLists.txt"))
        cmake_file.append(CMakeMinimumVersion("3.20"))
        cmake_file.append(CMakeProject(self.variant_name))
        cmake_file.append(CMakeInclude(self.variant_cmake_file))
        return cmake_file

    def create_variants_cmake(self) -> CMakeFile:
        cmake_file = CMakeFile(self.variant_cmake_file.to_path())
        vars = {"CMAKE_CXX_STANDARD": "99", "CMAKE_C_COMPILER": "clang", "CMAKE_CXX_COMPILER": "clang++"}
        for var, value in vars.items():
            cmake_file.append(CMakeVariable(var, value))
        cmake_file.append(CMakeInclude(self.components_cmake_file))
        cmake_file.append(self.get_include_directories())
        # TODO: I do not like that I have to know here that the components are object libraries
        variant_executable = CMakeAddExecutable(
            "${PROJECT_NAME}",
            sources=[],
            libraries=[CMakeObjectLibrary(component.name).target_name for component in self.environment.components],
        )

        cmake_file.append(variant_executable)
        cmake_file.append(
            CMakeCustomTarget(
                BuildVariantRequest(self.variant_name).target_name,
                f"Build variant {self.variant_name}",
                [],
                [variant_executable.name],
            )
        )
        if BuildComponentAnalyzer(self.environment.components).is_testable():
            # TODO: this is a command for calling a cmake runner.
            # Use the CMakeRunner class somehow otherwise the command options will be duplicated
            command = CMakeCommand(
                "${CMAKE_COMMAND}",
                [
                    "-S",
                    self.cmake_current_list_dir.joinpath("test"),
                    "-B",
                    self.cmake_current_list_dir.joinpath("test/build"),
                    "-G",
                    "Ninja",
                ],
            )
            cmake_file.append(CMakeExecuteProcess("Configure the test cmake project", [command]))
            test_variant_target = TestVariantRequest(self.variant_name).target_name
            command = CMakeCommand(
                "${CMAKE_COMMAND}",
                [
                    "--build",
                    self.cmake_current_list_dir.joinpath("test/build"),
                    "--target",
                    test_variant_target,
                ],
            )
            cmake_file.append(
                CMakeCustomTarget(
                    test_variant_target,
                    "Running all tests",
                    [command],
                )
            )
        else:
            cmake_file.append(CMakeComment("Skip including test project. No testable components found."))
        return cmake_file

    def get_include_directories(self) -> CMakeIncludeDirectories:
        collector = BuildComponentAnalyzer(self.environment.components)
        include_dirs = collector.collect_include_directories() + self.environment.include_directories
        return CMakeIncludeDirectories([CMakePath(path) for path in include_dirs])

    def create_components_cmake(self) -> CMakeFile:
        cmake_file = CMakeFile(self.components_cmake_file.to_path())
        for component in self.environment.components:
            component_analyzer = BuildComponentAnalyzer([component])
            component_library = CMakeObjectLibrary(component.name, component_analyzer.collect_sources())
            cmake_file.append(component_library)
            cmake_file.append(
                CMakeCustomTarget(
                    CompileComponentRequest(self.variant_name, component.name).target_name,
                    f"Compile component {component.name}",
                    [],
                    [component_library.target_name],
                )
            )
            if component_analyzer.is_testable():
                command = CMakeCommand(
                    "${CMAKE_COMMAND}",
                    [
                        "--build",
                        self.cmake_current_list_dir.joinpath("test/build"),
                        "--target",
                        f"{component.name}_test",
                    ],
                )
                cmake_file.append(
                    CMakeCustomTarget(
                        f"{component.name}_test",
                        f"Running {component.name}_test",
                        [command],
                    )
                )
        return cmake_file


class CmakeTestFilesGenerator:
    """Generates CMakeLists.txt, variant.cmake, and components.cmake files
    for running tests on the variant and its components."""

    def __init__(self, environment: BuildEnvironment, output_dir: Path) -> None:
        self.environment = environment
        self.output_dir = output_dir
        # The directory where the CMakeLists.txt file is located
        self.cmake_source_dir = CMakePath(self.output_dir, "CMAKE_CURRENT_LIST_DIR")
        # The directory where the build files will be generated
        self.cmake_build_dir = CMakePath(self.output_dir, "CMAKE_BUILD_DIR", Path("build"))

    @property
    def variant_name(self) -> str:
        return self.environment.variant_name

    @property
    def variant_cmake_file(self) -> CMakePath:
        return self.cmake_source_dir.joinpath("variant.cmake")

    @property
    def components_cmake_file(self) -> CMakePath:
        return self.cmake_source_dir.joinpath("components.cmake")

    def generate(self) -> List[CMakeFile]:
        return [self.create_cmake_lists(), self.create_variants_cmake(), self.create_components_cmake()]

    def create_cmake_lists(self) -> CMakeFile:
        cmake_file = CMakeFile(self.output_dir.joinpath("CMakeLists.txt"))
        cmake_file.append(CMakeMinimumVersion("3.20"))
        cmake_file.append(CMakeProject(self.variant_name))
        cmake_file.append(CMakeInclude(self.variant_cmake_file))
        return cmake_file

    def create_variants_cmake(self) -> CMakeFile:
        cmake_file = CMakeFile(self.variant_cmake_file.to_path())
        cmake_file.append(CMakeVariable("CMAKE_CXX_STANDARD", "14"))
        cmake_file.append(CMakeVariable("CMAKE_CXX_STANDARD_REQUIRED", "ON"))
        cmake_file.append(CMakeVariable("gtest_force_shared_crt", "ON", True, "BOOL", "", True))
        cmake_file.append(CMakeComment("Add local GoogleTest directory"))
        cmake_file.append(
            CMakeAddSubdirectory(
                self.cmake_source_dir.joinpath("../../../external/gtest"),
                self.cmake_build_dir.joinpath(".gtest"),
            )
        )
        cmake_file.append(CMakeInclude("GoogleTest"))
        cmake_file.append(CMakeComment("Enable testing"))
        cmake_file.append(CMakeInclude("CTest"))
        cmake_file.append(CMakeListAppend("CMAKE_CTEST_ARGUMENTS", ["--output-on-failure"]))
        cmake_file.append(CMakeEnableTesting())
        cmake_file.append(CMakeInclude(self.components_cmake_file))
        cmake_file.append(self.get_include_directories())
        return cmake_file

    def get_include_directories(self) -> CMakeIncludeDirectories:
        collector = BuildComponentAnalyzer(self.environment.components)
        include_dirs = collector.collect_include_directories() + self.environment.include_directories
        return CMakeIncludeDirectories([CMakePath(path) for path in include_dirs])

    def create_components_cmake(self) -> CMakeFile:
        cmake_file = CMakeFile(self.components_cmake_file.to_path())
        for component in self.environment.components:
            component_analyzer = BuildComponentAnalyzer([component])
            # Skip components without tests
            if not component_analyzer.is_testable():
                continue
            cmake_file.append(CMakeComment(f"Component {component.name}"))
            sources = component_analyzer.collect_sources() + component_analyzer.collect_test_sources()
            component_test_executable_target = f"{component.name}_build_test_executable"
            cmake_file.append(
                CMakeAddExecutable(
                    component_test_executable_target,
                    [CMakePath(source) for source in sources],
                    ["GTest::gtest_main"],
                )
            )
            command = CMakeCommand(
                "${CMAKE_CTEST_COMMAND}",
                [
                    "${CMAKE_CTEST_ARGUMENTS}",
                    "--output-junit",
                    f"{component.name}_junit.xml",
                    "||",
                    "${CMAKE_COMMAND}",
                    "-E",
                    "true",
                ],
            )
            outputs = [self.cmake_build_dir.joinpath(f"{component.name}_junit.xml")]
            cmake_file.append(
                CMakeCustomCommand(
                    "Run the test executable, generate JUnit report and return success independent of the test result",
                    outputs,
                    [component_test_executable_target],
                    [command],
                )
            )
            cmake_file.append(
                CMakeCustomTarget(
                    TestComponentRequest(self.variant_name, component.name).target_name,
                    f"Execute tests for {component.name}",
                    [],
                    outputs,  # type: ignore
                    True,
                )
            )
            cmake_file.append(CMakeContent(f"gtest_discover_tests({component_test_executable_target})"))
            cmake_file.append(CMakeEmptyLine())
        return cmake_file



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import Any, List, Optional, Union

from py_app_dev.core.logging import logger, time_it
from py_app_dev.core.subprocess import SubprocessExecutor

from yanga.ybuild.components import BuildComponent

from .builder import Builder
from .generated_file import GeneratedFile


def make_list_unique(seq: List[Any]) -> List[Any]:
    return list(dict.fromkeys(seq))


class LibraryType(Enum):
    OBJECT = auto()


class CMakeElement(ABC):
    tab_prefix = " " * 4

    @abstractmethod
    def to_string(self) -> str:
        pass

    def __str__(self) -> str:
        return self.to_string()


class CMakeContent(CMakeElement):
    def __init__(self, content: str) -> None:
        super().__init__()
        self.content = content

    def to_string(self) -> str:
        return self.content


class CMakeEmptyLine(CMakeElement):
    def to_string(self) -> str:
        return ""


class CMakeComment(CMakeElement):
    def __init__(self, comment: str) -> None:
        super().__init__()
        self.comment = comment

    def to_string(self) -> str:
        return f"# {self.comment}"


class CMakeProject(CMakeElement):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def to_string(self) -> str:
        return f"project({self.name})"


class CMakeMinimumVersion(CMakeElement):
    def __init__(self, version: str) -> None:
        super().__init__()
        self.version = version

    def to_string(self) -> str:
        return f"cmake_minimum_required(VERSION {self.version})"


@dataclass
class CMakeLibrary(CMakeElement):
    name: str
    files: List[Path] = field(default_factory=list)
    type: LibraryType = LibraryType.OBJECT

    @property
    def target_name(self) -> str:
        return f"{self.name}_lib"

    def to_string(self) -> str:
        return f"add_library({self.target_name} {self.type.name} {self._get_files_string()})"

    def _get_files_string(self) -> str:
        return " ".join([file.as_posix() for file in self.files])


class CMakeObjectLibrary(CMakeLibrary):
    def __init__(self, name: str, files: List[Path] = field(default_factory=list)) -> None:
        super().__init__(name, files, LibraryType.OBJECT)


@dataclass
class CMakeVariable(CMakeElement):
    """
    set(<variable> <value>... CACHE <type> <docstring> [FORCE])
    https://cmake.org/cmake/help/latest/command/set.html
    """

    name: str
    value: str
    cache: bool = False
    type: Optional[str] = None
    docstring: Optional[str] = None
    force: bool = False

    def to_string(self) -> str:
        arguments = [self.name, self.value]
        if self.cache:
            arguments.append("CACHE")
        if self.type:
            arguments.append(self.type)
        if self.docstring is not None:
            if self.docstring:
                arguments.append(self.docstring)
            else:
                arguments.append('""')
        if self.force:
            arguments.append("FORCE")
        return "set(" + " ".join(arguments) + ")"


class CMakePath:
    def __init__(
        self,
        path: Path,
        variable: Optional[str] = None,
        relative_path: Optional[Path] = None,
    ) -> None:
        super().__init__()
        self.path = path
        self.variable = variable
        self.relative_path = relative_path

    def to_cmake_element(self) -> Optional[CMakeElement]:
        return CMakeVariable(self.variable, self.to_string()) if self.variable else None

    def to_string(self) -> str:
        if self.variable:
            path = f"${{{self.variable}}}"
        else:
            path = self.path.as_posix()
        if self.relative_path:
            path += f"/{self.relative_path.as_posix()}"
        return path

    def to_path(self) -> Path:
        return self.path / self.relative_path if self.relative_path else self.path

    def joinpath(self, path: str) -> "CMakePath":
        return CMakePath(self.path, self.variable, Path(path))

    def __str__(self) -> str:
        return self.to_string()


class CMakeInclude(CMakeElement):
    def __init__(self, path: str | CMakePath) -> None:
        super().__init__()
        self.path = path

    def to_string(self) -> str:
        return f"include({self.path})"


class CMakeIncludeDirectories(CMakeElement):
    def __init__(self, paths: List[CMakePath]) -> None:
        super().__init__()
        self.paths = paths

    def to_string(self) -> str:
        return "\n".join(["include_directories("] + self._add_tabulated_paths(self.paths) + [")"])

    def _add_tabulated_paths(self, paths: List[CMakePath]) -> List[str]:
        return [self._add_tabulated_path(path) for path in paths]

    def _add_tabulated_path(self, path: CMakePath) -> str:
        return self.tab_prefix + path.to_string()


@dataclass
class CMakeAddExecutable(CMakeElement):
    """
    add_executable(<name> [WIN32] [MACOSX_BUNDLE]
                     [EXCLUDE_FROM_ALL]
                        [source1] [source2 ...])
    https://cmake.org/cmake/help/latest/command/add_executable.html#add-executable
    """

    name: str
    sources: List[Union[str, CMakePath, CMakeObjectLibrary]]
    libraries: List[str] = field(default_factory=list)
    exclude_from_all: bool = False

    def to_string(self) -> str:
        content = self._add_executable()
        if self.libraries:
            # add target_link_libraries
            content += "\n" + self._add_target_link_libraries()
        return content

    def _add_executable(self) -> str:
        arguments = [self.name]
        if self.exclude_from_all:
            arguments.append("EXCLUDE_FROM_ALL")
        arguments.extend(self._get_sources())
        return "add_executable(" + " ".join(arguments) + ")"

    def _add_target_link_libraries(self) -> str:
        return "target_link_libraries(" + " ".join([self.name] + self.libraries) + ")"

    def _get_sources(self) -> List[str]:
        return [self._get_source(source) for source in self.sources]

    def _get_source(self, source: str | CMakePath | CMakeObjectLibrary) -> str:
        """
        - str: return as is
        - CMakePath: return the path as string
        - CMakeLibrary: return the target objects
        """
        if isinstance(source, CMakeObjectLibrary):
            return f"$<TARGET_OBJECTS:{source.target_name}>"
        else:
            return str(source)


class CMakeCommand(CMakeElement):
    def __init__(self, command: str | CMakePath, arguments: List[str | CMakePath]) -> None:
        super().__init__()
        self.command = command
        self.arguments = arguments

    def to_string(self) -> str:
        return f"COMMAND {self.command} {' '.join(str(arg) for arg in self.arguments)}"


class CMakeExecuteProcess(CMakeElement):
    def __init__(self, description: str, commands: List[CMakeCommand]) -> None:
        super().__init__()
        self.description = description
        self.commands = commands

    def to_string(self) -> str:
        content = [CMakeComment(self.description), "execute_process("]
        content.extend(self._get_commands())
        content.extend(
            [
                f"{self.tab_prefix}RESULT_VARIABLE result",
                ")",
                "if(result)",
                f"{self.tab_prefix}message(FATAL_ERROR '{self.description} failed: ${{result}}')",
                "endif()",
            ]
        )
        return "\n".join(str(line) for line in content)

    def _get_commands(self) -> List[str]:
        return [self.tab_prefix + str(command) for command in self.commands]


class CMakeDepends(CMakeElement):
    def __init__(self, depends: List[str | CMakePath]) -> None:
        super().__init__()
        self.depends = depends

    def to_string(self) -> str:
        return f"{self.tab_prefix}DEPENDS {' '.join(str(depend) for depend in self.depends)}"


@dataclass
class CMakeCustomCommand(CMakeElement):
    description: str
    outputs: List[CMakePath]
    depends: List[str | CMakePath]
    commands: List[CMakeCommand]

    def to_string(self) -> str:
        content = [CMakeComment(self.description), "add_custom_command("]
        content.extend(self._get_outputs())
        content.append(CMakeDepends(self.depends).to_string())
        content.extend(self._get_commands())
        content.append(")")
        return "\n".join(str(line) for line in content)

    def _get_outputs(self) -> List[str]:
        return [f"{self.tab_prefix}OUTPUT {output.to_string()}" for output in self.outputs]

    def _get_commands(self) -> List[str]:
        return [self.tab_prefix + str(command) for command in self.commands]


@dataclass
class CMakeCustomTarget(CMakeElement):
    def __init__(
        self,
        name: str,
        description: str,
        commands: List[CMakeCommand],
        depends: Optional[List[str | CMakePath]] = None,
        default_target: bool = False,
    ) -> None:
        super().__init__()
        self.name = name
        self.description = description
        self.depends = depends
        self.commands = commands
        self.default_target = default_target

    def to_string(self) -> str:
        add_to_all_target = "ALL" if self.default_target else ""
        content = [
            CMakeComment(self.description),
            f"add_custom_target({self.name} {add_to_all_target}",
        ]
        content.extend(self._get_commands())
        if self.depends:
            content.append(CMakeDepends(self.depends).to_string())
        content.append(")")
        return "\n".join(str(line) for line in content)

    def _get_commands(self) -> List[str]:
        return [self.tab_prefix + str(command) for command in self.commands]


class CMakeAddSubdirectory(CMakeElement):
    """
    add_subdirectory(source_dir [binary_dir] [EXCLUDE_FROM_ALL] [SYSTEM])
    https://cmake.org/cmake/help/latest/command/add_subdirectory.html#add-subdirectory
    """

    def __init__(self, source_dir: CMakePath, build_dir: Optional[CMakePath] = None) -> None:
        super().__init__()
        self.source_dir = source_dir
        self.build_dir = build_dir

    def to_string(self) -> str:
        arguments = [self.source_dir.to_string()]
        if self.build_dir:
            arguments.append(self.build_dir.to_string())
        return "add_subdirectory(" + " ".join(arguments) + ")"


class CMakeListAppend(CMakeElement):
    def __init__(self, variable: str, values: List[str]) -> None:
        super().__init__()
        self.variable = variable
        self.values = values

    def to_string(self) -> str:
        return f"list(APPEND {self.variable} {self._get_values_string()})"

    def _get_values_string(self) -> str:
        return " ".join(self.values)


class CMakeFile(GeneratedFile):
    def __init__(self, path: Path) -> None:
        super().__init__(path)
        self.content: List[CMakeElement] = []

    def to_string(self) -> str:
        return "\n".join(str(elem) for elem in self.content)

    def append(self, content: Optional[CMakeElement]) -> "CMakeFile":
        if content:
            self.content.append(content)
        return self


class CMakeEnableTesting(CMakeElement):
    def __init__(self) -> None:
        super().__init__()

    def to_string(self) -> str:
        return "enable_testing()"


class CMakeLists(GeneratedFile):
    tab_prefix = " " * 4

    def __init__(self, path: Path) -> None:
        super().__init__(path)
        self.project_name = ""
        self.cmake_version = ""
        self.source_files: List[Path] = []
        self.include_directories: List[Path] = []
        self.libraries: List[CMakeLibrary] = []

    def to_string(self) -> str:
        content = [
            f"cmake_minimum_required(VERSION {self.cmake_version})",
            f"project({self.project_name})",
            "",
            "set(CMAKE_CXX_STANDARD 99)",
            "",
            "set(CMAKE_C_COMPILER clang)",
            "set(CMAKE_CXX_COMPILER clang++)",
            "",
        ]
        content.extend(self._generate_source_files())
        content.append("")
        content.extend(self._generate_libraries())
        content.append("")
        content.extend(self._generate_include_directories())
        libraries_deps = ""
        if self.libraries:
            libraries_deps = " ".join([f"$<TARGET_OBJECTS:{library.target_name}>" for library in self.libraries])

        content.append("")
        content.append(f"add_executable(${{PROJECT_NAME}} ${{SOURCE_FILES}} {libraries_deps})")
        return "\n".join(content) + "\n"

    def _generate_source_files(self) -> List[str]:
        return ["set(SOURCE_FILES "] + self._add_tabulated_paths(self.source_files) + [")"]

    def _generate_libraries(self) -> List[str]:
        return [library.to_string() for library in self.libraries]

    def _generate_include_directories(self) -> List[str]:
        return ["include_directories("] + self._add_tabulated_paths(self.include_directories) + [")"]

    def _add_tabulated_paths(self, paths: List[Path]) -> List[str]:
        return [self._add_tabulated_path(path) for path in paths]

    def _add_tabulated_path(self, path: Path) -> str:
        return self.tab_prefix + path.absolute().as_posix()


class BuildFileCollector:
    def __init__(self, components: List[BuildComponent]) -> None:
        self.components = components

    def collect_sources(self) -> List[Path]:
        files: List[Path] = []
        for component in self.components:
            files.extend([component.path.joinpath(source) for source in component.sources])
        return files

    def collect_include_directories(self) -> List[Path]:
        # TODO: check if there are specific include directories for each component
        includes = [path.parent for path in self.collect_sources()]
        # remove duplicates and return
        return make_list_unique(includes)


class CMakeListsBuilder(Builder):
    file_name = "CMakeLists.txt"

    def __init__(self, output_path: Path, cmake_version: str = "3.20") -> None:
        self.output_path = output_path
        self.components: List[BuildComponent] = []
        self.cmake_lists = CMakeLists(self.output_path.joinpath(self.file_name))
        self.cmake_lists.cmake_version = cmake_version
        self.include_directories: List[Path] = []

    def with_project_name(self, name: str) -> Builder:
        self.cmake_lists.project_name = name
        return self

    def with_components(self, components: List[BuildComponent]) -> Builder:
        self.components = components
        return self

    def with_include_directories(self, include_directories: List[Path]) -> Builder:
        self.include_directories.extend(include_directories)
        return self

    def build(self) -> GeneratedFile:
        collector = BuildFileCollector(self.components)
        for component in self.components:
            self.cmake_lists.libraries.append(
                CMakeLibrary(component.name, BuildFileCollector([component]).collect_sources())
            )
        self.cmake_lists.include_directories = make_list_unique(
            self.include_directories + collector.collect_include_directories()
        )
        return self.cmake_lists


class CMakeRunner:
    executable = "cmake"

    def __init__(self, install_directories: List[Path]) -> None:
        self.logger = logger.bind()
        self.install_directories = install_directories

    def run(self, build_dir: Path, target: str = "all") -> None:
        self.configure(build_dir)
        self.build(build_dir, target)

    @time_it("CMake configure")
    def configure(self, build_dir: Path) -> None:
        build_dir_str = build_dir.absolute().as_posix()
        arguments = f" -S{build_dir_str}" f" -B{build_dir_str}" f" -G Ninja "
        self.run_cmake(arguments)

    @time_it("CMake build")
    def build(self, build_dir: Path, target: str = "all") -> None:
        build_dir_str = build_dir.absolute().as_posix()
        arguments = f" --build {build_dir_str}" f" --target {target} -- "
        self.run_cmake(arguments)

    def run_cmake(self, arguments: str) -> None:
        # Add the install directories to the PATH
        env = os.environ.copy()
        env["PATH"] = ";".join([path.absolute().as_posix() for path in self.install_directories] + [env["PATH"]])
        command = self.executable + " " + arguments
        SubprocessExecutor([command], env=env).execute()



from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from mashumaro import DataClassDictMixin
from pick import pick
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger, time_it

from yanga.project.project_slurper import YangaProjectSlurper
from yanga.ybuild.environment import BuildEnvironment
from yanga.ybuild.generators.build_system_request import (
    BuildSystemRequest,
    CustomBuildSystemRequest,
)
from yanga.ybuild.pipeline import StageRunner


@dataclass
class BuildCommandConfig(DataClassDictMixin):
    # If the user does not specify the variant name, it shall be prompted for it.
    variant_name: Optional[str] = field(default=None, metadata={"help": "SPL variant name."})
    project_dir: Path = field(
        default=Path(".").absolute(),
        metadata={"help": "Project root directory. " "Defaults to the current directory if not specified."},
    )
    target: Optional[str] = field(default=None, metadata={"help": "Target to build."})

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "BuildCommandConfig":
        return cls.from_dict(vars(namespace))


class BuildCommand(Command):
    def __init__(self) -> None:
        super().__init__("build", "Build a yanga project")
        self.logger = logger.bind()

    @time_it("Build")
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        return self.do_run(BuildCommandConfig.from_namespace(args))

    def do_run(self, config: BuildCommandConfig) -> int:
        project = YangaProjectSlurper(config.project_dir)
        variant_name = self.select_variant(project, config.variant_name)
        if not variant_name:
            raise UserNotificationException("No variant selected.")
        if config.target:
            build_request = CustomBuildSystemRequest(variant_name, config.target)
        else:
            # TODO: I have no idea why mypy complains here.
            # BuildSystemRequest is a base class for CustomBuildSystemRequest.
            build_request = BuildSystemRequest(variant_name)  # type: ignore
        build_environment = BuildEnvironment(
            config.project_dir,
            build_request,
            project.get_variant_components(variant_name),
            project.user_config_files,
            project.get_variant_config_file(variant_name),
        )
        for stage in project.stages:
            StageRunner(build_environment, stage).run()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, BuildCommandConfig)

    def select_variant(self, project: YangaProjectSlurper, variant_name: Optional[str]) -> Optional[str]:
        if variant_name:
            return variant_name
        if not project.variants:
            return None
        variants = [variant.name for variant in project.variants]
        try:
            # TODO: this message is only necessary in case the user will press Ctrl+C to quit.
            # In this case, after pick method returns, the execution is paused until the user presses any key.
            # I have no idea why that happens.
            self.logger.info("Press any key to continue...")
            selected_variant, _ = pick(variants, "Select a variant: ", indicator="=>")
        except KeyboardInterrupt:
            selected_variant = None
        return str(selected_variant) if selected_variant else None



from enum import auto
from pathlib import Path
from typing import List, Optional

import customtkinter
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger, time_it
from py_app_dev.core.subprocess import SubprocessExecutor
from py_app_dev.mvp.event_manager import EventID, EventManager
from py_app_dev.mvp.presenter import Presenter
from py_app_dev.mvp.view import View

from yanga.project.project_slurper import YangaProjectSlurper
from yanga.ybuild.environment import BuildEnvironment
from yanga.ybuild.generators.build_system_request import (
    BuildSystemRequest,
    BuildVariantRequest,
    CleanVariantRequest,
    CompileComponentRequest,
    TestComponentRequest,
)
from yanga.ybuild.pipeline import StageRunner

from .icons import Icons


class YangaEvent(EventID):
    BUILD_EVENT = auto()
    COMPONENT_COMPILE_EVENT = auto()
    COMPONENT_TEST_EVENT = auto()
    REFRESH_EVENT = auto()
    VARIANT_SELECTED_EVENT = auto()
    CLEAN_VARIANT_EVENT = auto()
    OPEN_IN_VSCODE = auto()


class YangaView(View):
    def __init__(self, event_manager: EventManager) -> None:
        self.event_manager = event_manager
        self.root = customtkinter.CTk()
        self.variants: List[str] = []
        self.components: List[str] = []

    @property
    def selected_variant(self) -> str:
        return self.variant_selection.get()

    @property
    def selected_component(self) -> str:
        return self.component_selection.get()

    def init_gui(self) -> None:
        customtkinter.set_default_color_theme("green")

        # Configure the main window
        self.root.title("YANGA")
        self.root.geometry(f"{220}x{400}")

        # update app icon
        self.root.iconbitmap(Icons.YANGA_ICON.file)
        position_in_grid = 0

        self.variants_frame = self._create_variants_frame(self.root, position_in_grid)
        position_in_grid += 1
        self.components_frame = self._create_components_frame(self.root, position_in_grid)

        # Bind F5 key to refresh functionality
        self.root.bind("<F5>", self._refresh_button_pressed)

        # Create events
        self.build_trigger = self.event_manager.create_event_trigger(YangaEvent.BUILD_EVENT)
        self.clean_variant_trigger = self.event_manager.create_event_trigger(YangaEvent.CLEAN_VARIANT_EVENT)
        self.component_compile_trigger = self.event_manager.create_event_trigger(YangaEvent.COMPONENT_COMPILE_EVENT)
        self.component_test_trigger = self.event_manager.create_event_trigger(YangaEvent.COMPONENT_TEST_EVENT)
        self.refresh_trigger = self.event_manager.create_event_trigger(YangaEvent.REFRESH_EVENT)
        self.variant_selected_trigger = self.event_manager.create_event_trigger(YangaEvent.VARIANT_SELECTED_EVENT)
        self.open_in_vscode_trigger = self.event_manager.create_event_trigger(YangaEvent.OPEN_IN_VSCODE)

    def _build_button_pressed(self) -> None:
        self.build_trigger(self.selected_variant)

    def _refresh_button_pressed(self, event=None) -> None:  # type: ignore
        self.refresh_trigger()

    def _variant_selected(self, variant_selected: str) -> None:
        self.variant_selected_trigger(variant_selected)

    def _component_compile_button_pressed(self) -> None:
        self.component_compile_trigger(self.selected_variant, self.selected_component)

    def _component_test_button_pressed(self) -> None:
        self.component_test_trigger(self.selected_variant, self.selected_component)

    def _clean_variant_button_pressed(self) -> None:
        self.clean_variant_trigger(self.selected_variant)

    def _open_in_vscode_button_pressed(self) -> None:
        self.open_in_vscode_trigger()

    def mainloop(self) -> None:
        self.root.mainloop()

    def _create_variants_frame(self, root: customtkinter.CTk, position_in_root_grid: int) -> customtkinter.CTkFrame:
        # Create the frame for all elements related to variants
        variants_frame = customtkinter.CTkFrame(root)
        variants_frame.grid(row=position_in_root_grid, column=0, sticky="nsew", padx=10, pady=10)
        current_frame = variants_frame
        position_in_grid = 0

        # Create label aligned to left
        variants_label = customtkinter.CTkLabel(current_frame, text="Variants", anchor="w")
        variants_label.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        # Create selection list
        self.variant_selection = customtkinter.CTkOptionMenu(current_frame, command=self._variant_selected)
        self.variant_selection.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        self.update_variants(self.variants)
        position_in_grid += 1

        # Create the build button
        self.build_button = customtkinter.CTkButton(current_frame, text="Build", command=self._build_button_pressed)
        self.build_button.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        # Create the clean button
        self.clean_button = customtkinter.CTkButton(
            current_frame, text="Clean", command=self._clean_variant_button_pressed
        )
        self.clean_button.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        # Create the open in vscode button
        self.open_in_vscode_button = customtkinter.CTkButton(
            current_frame, text="Open in VSCode", command=self._open_in_vscode_button_pressed
        )
        self.open_in_vscode_button.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        return variants_frame

    def _create_components_frame(self, root: customtkinter.CTk, position_in_root_grid: int) -> customtkinter.CTkFrame:
        # Create the frame for all elements related to components
        components_frame = customtkinter.CTkFrame(root)
        components_frame.grid(row=position_in_root_grid, column=0, sticky="nsew", padx=10, pady=10)
        current_frame = components_frame
        position_in_grid = 0

        # Create label aligned to left
        components_label = customtkinter.CTkLabel(current_frame, text="Components", anchor="w")
        components_label.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        # Create selection list
        self.component_selection = customtkinter.CTkOptionMenu(current_frame)
        self.component_selection.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)

        position_in_grid += 1

        # Create component compile button
        self.component_compile_button = customtkinter.CTkButton(
            current_frame, text="Compile", command=self._component_compile_button_pressed
        )
        self.component_compile_button.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        # Create component test button
        self.component_test_button = customtkinter.CTkButton(
            current_frame, text="Test", command=self._component_test_button_pressed
        )
        self.component_test_button.grid(row=position_in_grid, column=0, sticky="nsew", padx=10, pady=5)
        position_in_grid += 1

        return components_frame

    def update_variants(self, variants: List[str]) -> None:
        self.variant_selection.configure(values=variants)

    def update_current_variant(self, variant: str) -> None:
        self.variant_selection.set(variant)

    def enable_variant_commands(self) -> None:
        self.build_button.configure(state="normal")
        self.clean_button.configure(state="normal")

    def disable_variant_commands(self) -> None:
        self.build_button.configure(state="disabled")
        self.clean_button.configure(state="disabled")

    def update_components(self, components: List[str]) -> None:
        self.component_selection.configure(values=components)

    def update_current_component(self, component: str) -> None:
        self.component_selection.set(component)

    def enable_component_commands(self) -> None:
        self.component_compile_button.configure(state="normal")

    def disable_component_commands(self) -> None:
        self.component_compile_button.configure(state="disabled")


class YangaPresenter(Presenter):
    def __init__(self, view: YangaView, event_manager: EventManager, project_dir: Path) -> None:
        self.logger = logger.bind()
        self.view = view
        self.event_manager = event_manager
        self.project_dir = project_dir
        self.project = self._create_project()
        self.event_manager.subscribe(YangaEvent.BUILD_EVENT, self._build_trigger)
        self.event_manager.subscribe(YangaEvent.COMPONENT_COMPILE_EVENT, self._component_compile_trigger)
        self.event_manager.subscribe(YangaEvent.COMPONENT_TEST_EVENT, self._test_compile_trigger)
        self.event_manager.subscribe(YangaEvent.REFRESH_EVENT, self._refresh_trigger)
        self.event_manager.subscribe(YangaEvent.VARIANT_SELECTED_EVENT, self._variant_selected_trigger)
        self.event_manager.subscribe(YangaEvent.CLEAN_VARIANT_EVENT, self._clean_variant_trigger)
        self.event_manager.subscribe(YangaEvent.OPEN_IN_VSCODE, self._open_in_vscode_trigger)
        self.command_running_flag = False
        self.running_build_system_request: Optional[BuildSystemRequest] = None
        self.selected_variant: Optional[str] = None
        self.selected_component: Optional[str] = None

    def run(self) -> None:
        self.view.init_gui()
        self._update_view_data()
        self.view.mainloop()

    def _build_trigger(self, variant_name: str) -> None:
        self.run_command(BuildVariantRequest(variant_name))

    def _component_compile_trigger(self, variant_name: str, component_name: str) -> None:
        self.run_command(CompileComponentRequest(variant_name, component_name))

    def _test_compile_trigger(self, variant_name: str, component_name: str) -> None:
        self.run_command(TestComponentRequest(variant_name, component_name))

    def _refresh_trigger(self) -> None:
        self.project = self._create_project()
        self._update_view_data()

    def _variant_selected_trigger(self, variant_name: str) -> None:
        self.logger.info(f"Variant selected: {variant_name}")
        self.selected_variant = variant_name
        self._update_components()

    def _clean_variant_trigger(self, variant_name: str) -> None:
        self.run_command(CleanVariantRequest(variant_name))

    def _open_in_vscode_trigger(self) -> None:
        if not self.project:
            self.logger.warning("Project is not loaded")
            return
        self.logger.info("Opening project in VSCode")
        try:
            SubprocessExecutor(["code", self.project_dir.as_posix()], shell=True).execute()  # nosec
        except UserNotificationException as e:
            self.logger.error(e)

    def _update_view_data(self) -> None:
        self._update_variants()
        self._update_components()

    def _update_variants(self) -> None:
        variants = []
        if self.project:
            variants = [variant.name for variant in self.project.variants]
        if variants:
            variants.sort()
            self.selected_variant = variants[0]
        else:
            variants = ["No variants found"]
            self.selected_variant = None
        self.view.update_variants(variants)
        self.view.update_current_variant(variants[0])
        if self.selected_variant:
            self.view.enable_variant_commands()
        else:
            self.view.disable_variant_commands()

    def _update_components(self) -> None:
        components = []
        if self.project and self.selected_variant:
            components = self._create_component_names(self.selected_variant)
        if components:
            components.sort()
            self.selected_component = components[0]
        else:
            components = ["No components found"]
            self.selected_component = None
        self.view.update_components(components)
        self.view.update_current_component(components[0])
        if self.selected_component:
            self.view.enable_component_commands()
        else:
            self.view.disable_component_commands()

    @time_it("executing command")
    def run_command(self, build_system_request: BuildSystemRequest) -> None:
        # Make sure the project is loaded before running any command.
        # Otherwise, there might be configuration changes that are not reflected in the project.
        self.project = self._create_project()
        if not self.project:
            self.logger.warning("Project is not loaded")
            return
        if self.command_running_flag:
            self.logger.warning(
                f"Command '{self.running_build_system_request}' still running. Skip starting new command."
            )
            return
        if not self.selected_variant:
            UserNotificationException("No variant selected. This looks like a bug.")
        self.logger.info(f"Selected variant: {build_system_request.variant_name}")
        if build_system_request.component_name:
            self.logger.info(f"Selected component: {build_system_request.component_name}")
        self.logger.info(f"Selected command: {build_system_request.command}")
        self.command_running_flag = True
        self.running_build_system_request = build_system_request
        try:
            build_environment = BuildEnvironment(
                self.project_dir,
                build_system_request,
                self.project.get_variant_components(build_system_request.variant_name),
                self.project.user_config_files,
                self.project.get_variant_config_file(build_system_request.variant_name),
            )
            for stage in self.project.stages:
                StageRunner(build_environment, stage).run()
        except UserNotificationException as e:
            self.logger.error(e)
        finally:
            self.command_running_flag = False
            self.running_build_system_request = None

    def _create_project(self) -> Optional[YangaProjectSlurper]:
        try:
            return YangaProjectSlurper(self.project_dir)
        except UserNotificationException as e:
            self.logger.error(e)
            return None

    def _create_component_names(self, variant_name: str) -> List[str]:
        if self.project:
            try:
                components = self.project.get_variant_components(variant_name)
                return [component.name for component in components]
            except UserNotificationException as e:
                self.logger.error(e)
                return []
        return []


class YangaGui:
    def __init__(self, project_dir: Path) -> None:
        self.project_dir = project_dir
        self.logger = logger.bind()

    def run(self) -> None:
        try:
            event_manager = EventManager()
            view = YangaView(event_manager)
            presenter = YangaPresenter(view, event_manager, self.project_dir)
            presenter.run()
        except UserNotificationException as e:
            self.logger.error(e)



from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path

from mashumaro import DataClassDictMixin
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.logging import logger, time_it

from yanga.gui.ygui import YangaGui


@dataclass
class GuiCommandConfig(DataClassDictMixin):
    project_dir: Path = field(
        default=Path(".").absolute(),
        metadata={"help": "Project root directory. " "Defaults to the current directory if not specified."},
    )

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "GuiCommandConfig":
        return cls.from_dict(vars(namespace))


class GuiCommand(Command):
    def __init__(self) -> None:
        super().__init__("gui", "Start the GUI to build an SPL project.")
        self.logger = logger.bind()

    @time_it()
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = GuiCommandConfig.from_namespace(args)
        YangaGui(config.project_dir).run()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, GuiCommandConfig)
